/*----------------------------------------------------------------------------
FILE NAME:     rstream.cpp

AUTHOR:        Stefano Basagni

CREATED:       February 1999

DESCRIPTION: This file contains the implementation of a random
             stream. The random stream is obtained by using the
             linear congruential generator theory for which,
             starting from a modulus m > 0, a multiplier a >= 0
             and < m, an increment c >= 0 and < m and an odd
             seed X(0), a random stream is generated in the
             following way:
                            X(i+1)  =  X(i)*a +c(mod m).
             In our case: m = 2**(32), a = 663608941 and c = 0.
             The seed has to be given in input to the class
             constructor, thus obtaining several different random
             streams.

             NOTE: It is assumed that the arithmetic for the
                   unsigned long is modulus 2**(32).

---
Copyright (C) 2003 Stefano Basagni

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

To receive a copy of the GNU General Public License write to
the Free Software Foundation, Inc., 59 Temple Place - Suite
330, Boston, MA 02111-1307, USA.

Stefano Basagni can be contacted at basagni@ece.neu.edu.

----------------------------------------------------------------------------*/
#include <math.h>

#include "rstream.h"

const double PII = 3.141592654;

const unsigned long int FACTOR = 663608941;

const double max_value = 4294967295.0; // (2**(32))-1

// This is used in the lngamma function
const float cof[6] = {76.18009173, -86.50532033, 24.01409822,
                      -1.231739516, 0.120858003e-2, -0.536382e-5};


/***********************************************************************
** function:    Rstream, constructor;
** parameters:  unsigned long int, a seed s;
** description: Create a stream with a seed s.
***********************************************************************/
Rstream::Rstream( unsigned long int s ) {

  xi = ( s % 2 ) ? s :( s + 1 ); // Make sure the seed is an odd number

} // End constructor


/***********************************************************************
** function:    uniform01;
** parameters:  none;
** description: Generate a random number between 0 and 1 based on the
**              relation X(i+1)  =  (X(i) * 663608941) mod (2**32).
***********************************************************************/
double Rstream::uniform01() {

  xi *= FACTOR;
  // This includes the '% 2**32', since the g++ arithmetic is % 2**32

  return ( (double)xi / max_value );

} // End uniform01


/***********************************************************************
** function:    uniform;
** parameters:  double, lower and upper;
** description: Generate a random number between lower and upper.
***********************************************************************/
double Rstream::uniform( double lower, double upper ) {

  return ( uniform01() * (upper - lower) + lower );

} // End uniform


/***********************************************************************
** function:    lngamma;
** parameters:  float, xx;
** description: Return log of gamma function of xx for xx > 0.
***********************************************************************/
float Rstream::lngamma( float  xx ) const {

  float  x, tmp, ser;
  int j;

  x = xx - 1.0;
  tmp = x + 5.5;
  tmp -= ( x + 0.5 ) * log( tmp );
  ser = 1.0;
  for ( j = 0; j < 6; j++ ) {
         x += 1.0;
         ser += cof[ j ] / x;
  }

  return ( -tmp + log( 2.50662827465 * ser ) );

} // End lngamma


/***********************************************************************
** function:    poisson;
** parameters:  float, mean;
** description: Return a random variable drawn from the poisson
**              distribution with average value mean. The value returned 
**              represents the number of unit rate poisson events 
**              occurring.
***********************************************************************/
double Rstream::poisson( float mean ) {

  static float sq, alxm, g, oldm = -1.0;
  float  em, t, y;

  /*
  ** Can we use direct method
  */
  if ( mean < 12.0 ) {
    if ( mean != oldm ) {
      oldm = mean;
      g    = exp( -mean );
    }
    em = -1.0;
    t  = 1.0;
    do {
      em += 1.0;
      t *= uniform01();
    } while ( t > g );
  }
  /*
  ** If not, use rejection method
  */
  else {
    if ( mean != oldm ) {
      oldm = mean;
      sq   = sqrt( 2.0 * mean );
      alxm = log( mean );
      g    = mean * alxm - lngamma( mean + 1.0 );
    }
    do {
      do {
        /*
        ** y is a deviate from a Lorentzian comparison function
        */
        y    = tan ( PII * uniform01() );
        em   = sq * y + mean;
      } while ( em < 0.0 );
      /*
      ** poisson is integer valued
      */
      em = floor( em );
      /*
      ** t is the ration of the desired distribution to the comparison
      ** function - accept or reject by comparing it to the the uniform
      ** deviate. Factor 0.9 is chosen so that it never exceeds 1.0
      */
      t = 0.9 * ( 1.0 + y * y ) * exp( em * alxm - 
				       lngamma( em + 1.0 ) - g );
    } while ( uniform01() > t );
  }

  return em;

} // End poisson


/***********************************************************************
** function:    uniform_discrete
** parameters:  int, lower, upper
** description: Generate a discrete random number between lower and 
**              upper (lower and upper included).
***********************************************************************/
unsigned long int Rstream::uniform_discrete( unsigned long int lower, 
					     unsigned long int upper ) {

  return ( ( unsigned long int )floor( uniform01()*( upper - lower + 1 ) 
				       + lower ) );

} // End uniform_discrete
